class PgDmDesignSkillTailwindOptions extends PgDmDesignSkill {
    constructor(version, provider) {
        super('tailwind', 'Tailwind CSS options');

        this.version = version;
        this.provider = provider;

        var _this = this;

        this.openDialog = function() {
            var settings = provider.settings;
            var project = pinegrow.getCurrentProject();

            if(settings.jit && settings.jit === 'true') {
                settings.mode = 'jit';
            }

            if(settings.external_source_url) {
                settings.external_source_url = project.getAbsoluteUrl(settings.external_source_url);
            }

            var old_mode = settings.mode;

            var fields = {
                mode: {
                    name: 'Compiler',
                    type: 'select',
                    value: settings.mode,
                    default_value: 'jit',
                    options: function() {
                        var r = [
                            {key: 'jit', name: 'Built-in JIT compiler (3.2.1)'},
                            {key: 'jit2', name: 'Built-in JIT compiler (2.2.6)'},
                            {key: 'full', name: 'Built-in non-JIT compiler (2.2.6)'},
                        ]
                        isApp() && r.push({key: 'external', name: 'External build process'});
                        return r;
                    },
                    helptext: `<p>Choose how to compile the Tailwind CSS stylesheet. Options are:</p>
<ul>
<li><b>Built-in JIT compiler</b> - Use built-in Tailwind compiler to generate optimized Tailwind CSS with just CSS rules that are used in the project.</li>
<li><b>Built-in full (non-JIT) compiler</b> - Use built-in Tailwind compiler to generate full Tailwind CSS stylesheet. Not recommended due to huge size.</li>
<li><b>External build process</b> - Use an external build process with <code>npm</code> or standalone executable. This option gives you the most flexibility and it lets you use whatever Tailwind version you need. This option is only available in the desktop Pinegrow version.</li>
</ul>`
                },
                /*
                jit: {
                    name: 'Generate optimized CSS with just-in-time compiler (recommended)',
                    type: 'checkbox',
                    value: 'true',
                    negvalue: 'false',
                    default_value: settings.jit,
                    helptext: `<p>Check to generate optimized Tailwind CSS with just CSS rules that are used in the project, using the Tailwind CSS Just-in-time compiler. If unchecked, the full Tailwind CSS stylesheet is generated each time you make a design change.</p>`
                },*/
                external_source_url: {
                    name: 'Source CSS file',
                    type: 'text',
                    value: settings.external_source_url,
                    file_picker_use_project_as_parent: true,
                    file_picker : true,
                    file_picker_no_proxy : true,
                    show_if: 'mode==external',
                    validate: function(pgel, fn, value, fdef, $field, values) {
                        if(values.mode === 'external') {
                            if(!value) {
                                return 'Source CSS file is required.'
                            } else {
                                var url = project.getAbsoluteUrl(value);
                                if(!crsaIsFileExist(crsaMakeFileFromUrl(url))) {
                                    return 'The file does not exist.'
                                }
                            }
                        }
                    },
                    helptext: `<p>Select the source css file of your external build process. That is usually the css files that contains statements:</p>
<pre>
@tailwind base;
@tailwind components;
@tailwind utilities;
</pre>`
                },
                external_help: {
                    name: 'Learn more about using external build',
                    type: 'button',
                    show_if: 'mode==external',
                    func: function() {
                        pinegrow.openExternalUrl('https://pinegrow.com/docs/tailwind/using-external-build-process/')
                    }
                }
            }


            pinegrow.showQuickDialog('Tailwind CSS Options', `Options`, `<p>Decide how to compile the Tailwind CSS stylesheet.</p><p>Note, these settings are stored in the project settings and therefore apply to all designs in the project.</p>`, fields, 'Cancel', 'Save', null, function(values) {
                if(settings.external_source_url) {
                    settings.external_source_url = project.getRelativeUrl(settings.external_source_url);
                }


                if(old_mode !== values.mode && (old_mode === 'external' || values.mode === 'external')) {
                    provider.setSettings(values, true);
                    provider.recreateEngine(function() {
                        if(old_mode === 'external') {
                            pinegrow.getAllPages().forEach(function(page) {
                                provider.getEngine(function(engine) {
                                    engine.activateOnPage(page);
                                })
                            })
                        }
                        pgDmStore.update();
                    })

                    if(values.mode === 'external') {
                        pinegrow.showAlert(`<p>Please make sure that the externally compiled Tailwind CSS stylesheet is included on HTML pages in this project. If the built-in compiler was used before, you will also have to remove any references to <code>tailwind_theme/tailwind.css</code> from the page.</p><p>The easiest way to add the stylesheet is to drag it from the Project panel to the page view.</p><p>To remove the old stylesheet from the page, delete its style node in the Tree panel.</p>`, 'Add the Tailwind CSS stylesheet to the page, if needed', null, 'Got it');
                    }
                } else {
                    provider.setSettings(values);
                }
            })
        }

        if(isApp()) {
            this.addControl('button.options', {
                type: 'button',
                name: 'Compiler options...',
                func: function () {
                    _this.openDialog();
                }
            })
        }

        this.addControl('disable.preflight',
            {
                name: 'Disable preflight',
                type: 'checkbox',
                value: '1',
                helptext: `<p>Preflight is a collection of CSS rules that reset all default browser styling. Check this to disable it.</p>`
            }
        )
    }

    populateDefaultParams() {
        this.setParam('disable.preflight', '0', true);
    }
}

class PgDmDesignSkillColorsForTailwind extends PgDmDesignSkillColors {
    constructor(colors) {
        super('tw.colors', colors, function() {
            return '';
        });
    }

    onGetConfigObject(part, config) {
        if(part === 'init') {
            config.theme = config.theme || {};
            config.theme.extend = config.theme.extend || {}
            config.theme.extend.colors = config.theme.extend.colors || {};

            this.colors.forEachColor(function (key, val, color_object) {
                try {
                    if (color_object.hasShades()) {
                        var d = {}
                        color_object.forEachShade(function (key, val) {
                            d[key] = chroma(val).hex();
                        })
                        config.theme.extend.colors[key] = d;
                    } else {
                        config.theme.extend.colors[key] = chroma(val).hex();
                    }
                } catch(err) {
                    console.error(err);
                }
            })
        }
    }

    populateDefaultParams(keep_existing) {

        if(!keep_existing) this.colors.clear();

        this.colors.add('gray', 'rgba(107,114,128,1)', true, {
            '50': 'rgba(249,250,251,1)',
            '100': 'rgba(243,244,246,1)',
            '200': 'rgba(229,231,235,1)',
            '300': 'rgba(209,213,219,1)',
            '400': 'rgba(156,163,175,1)',
            '500': 'rgba(107,114,128,1)',
            '600': 'rgba(75,85,99,1)',
            '700': 'rgba(55,65,81,1)',
            '800': 'rgba(31,41,55,1)',
            '900': 'rgba(17,24,39,1)'
        }, true);
        this.colors.add('red', 'rgba(239,68,68,1)', true, {
            '50': 'rgba(254,242,242,1)',
            '100': 'rgba(254,226,226,1)',
            '200': 'rgba(254,202,202,1)',
            '300': 'rgba(252,165,165,1)',
            '400': 'rgba(248,113,113,1)',
            '500': 'rgba(239,68,68,1)',
            '600': 'rgba(220,38,38,1)',
            '700': 'rgba(185,28,28,1)',
            '800': 'rgba(153,27,27,1)',
            '900': 'rgba(127,29,29,1)'
        }, true);
        this.colors.add('yellow', 'rgba(245,158,11,1)', true, {
            '50': 'rgba(255,251,235,1)',
            '100': 'rgba(254,243,199,1)',
            '200': 'rgba(253,230,138,1)',
            '300': 'rgba(252,211,77,1)',
            '400': 'rgba(251,191,36,1)',
            '500': 'rgba(245,158,11,1)',
            '600': 'rgba(217,119,6,1)',
            '700': 'rgba(180,83,9,1)',
            '800': 'rgba(146,64,14,1)',
            '900': 'rgba(120,53,15,1)'
        }, true);
        this.colors.add('green', 'rgba(16,185,129,1)', true, {
            '50': 'rgba(236,253,245,1)',
            '100': 'rgba(209,250,229,1)',
            '200': 'rgba(167,243,208,1)',
            '300': 'rgba(110,231,183,1)',
            '400': 'rgba(52,211,153,1)',
            '500': 'rgba(16,185,129,1)',
            '600': 'rgba(5,150,105,1)',
            '700': 'rgba(4,120,87,1)',
            '800': 'rgba(6,95,70,1)',
            '900': 'rgba(6,78,59,1)'
        }, true);
        this.colors.add('blue', 'rgba(59,130,246,1)', true, {
            '50': 'rgba(239,246,255,1)',
            '100': 'rgba(219,234,254,1)',
            '200': 'rgba(191,219,254,1)',
            '300': 'rgba(147,197,253,1)',
            '400': 'rgba(96,165,250,1)',
            '500': 'rgba(59,130,246,1)',
            '600': 'rgba(37,99,235,1)',
            '700': 'rgba(29,78,216,1)',
            '800': 'rgba(30,64,175,1)',
            '900': 'rgba(30,58,138,1)'
        }, true);
        this.colors.add('indigo', 'rgba(99,102,241,1)', true, {
            '50': 'rgba(238,242,255,1)',
            '100': 'rgba(224,231,255,1)',
            '200': 'rgba(199,210,254,1)',
            '300': 'rgba(165,180,252,1)',
            '400': 'rgba(129,140,248,1)',
            '500': 'rgba(99,102,241,1)',
            '600': 'rgba(79,70,229,1)',
            '700': 'rgba(67,56,202,1)',
            '800': 'rgba(55,48,163,1)',
            '900': 'rgba(49,46,129,1)'
        }, true);
        this.colors.add('purple', 'rgba(139,92,246,1)', true, {
            '50': 'rgba(245,243,255,1)',
            '100': 'rgba(237,233,254,1)',
            '200': 'rgba(221,214,254,1)',
            '300': 'rgba(196,181,253,1)',
            '400': 'rgba(167,139,250,1)',
            '500': 'rgba(139,92,246,1)',
            '600': 'rgba(124,58,237,1)',
            '700': 'rgba(109,40,217,1)',
            '800': 'rgba(91,33,182,1)',
            '900': 'rgba(76,29,149,1)'
        }, true);
        this.colors.add('pink', 'rgba(236,72,153,1)', true, {
            '50': 'rgba(253,242,248,1)',
            '100': 'rgba(252,231,243,1)',
            '200': 'rgba(251,207,232,1)',
            '300': 'rgba(249,168,212,1)',
            '400': 'rgba(244,114,182,1)',
            '500': 'rgba(236,72,153,1)',
            '600': 'rgba(219,39,119,1)',
            '700': 'rgba(190,24,93,1)',
            '800': 'rgba(157,23,77,1)',
            '900': 'rgba(131,24,67,1)'
        }, true);

        super.populateDefaultParams();
    }


}
class PgDmDesignProviderForTailwind extends PgDmDesignProvider {

    constructor(f) {
        super('tw', 'Tailwind CSS Design');
        this.framework = f;

        var _this = this;

        this.on_engine_ready_funcs = []

        this.colors = new PgDmDesignColors();
        this.colors.can_edit = true;
        this.colors.mode = 'class';
        this.colors.default_shades = ['50', '100', '200', '300', '400', '500', '600', '700', '800', '900'];

        this.setDefaultSettings();

        this.addSkills();
    }

    useExternalBuild() {
        return isApp() && this.settings.mode === 'external';
    }

    useTailwindJIT() {
        return this.settings.mode === 'jit' || this.settings.mode === 'jit2';
        //return this.settings.jit === 'true';
    }

    disablePreflight() {
        return this.options_skill.getParam('disable.preflight') === '1';
    }

    getExternalConfig(url, error_value) {
        url = url || this.settings.config_url;
        if(!url) return {};
        if(!crsaIsAbsoluteUrl(url)) {
            var p = pinegrow.getCurrentProject();
            url = p.getAbsoluteUrl(url);
        }
        try {
            var file = crsaMakeFileFromUrl(url);
            var code = require('fs').readFileSync(file, {encoding: 'utf8'});
            var module = Function('var module = {}; return (' + code + ')')();
            //console.log('Got external config', module);
            return module;
        } catch(err) {
            //console.error(err);
            return error_value !== undefined ? error_value : {}
        }
    }

    getTailwindVersion() {
        switch(this.settings.mode) {
            case 'jit':
                return 3.0;
            case 'jit2':
            case 'full':
                return 2.2;
            default:
                return 3.0;
        }
    }

    activate() {
        super.activate();
    }


    firstActivateInProject(done) {
        var _this = this;

        //reset settings
        _this.default_compiler = 'jit';
        _this.setDefaultSettings();

        var project = pinegrow.getCurrentProject();
        if(project.findFilesByName('tailwind.config.js').length) {
            pinegrow.showAlert(`<p>The project contains the <code>tailwind.config.js</code> file. If you are using an existing Tailwind CSS build process outside of Pinegrow, you can keep using it with the Design panel, after making <a href="https://pinegrow.com/docs/tailwind/using-external-build-process/" class="external">a couple of simple changes</a>.</p><p>Which Tailwind CSS compiler would you like to use?</p>
<ul>
<li>The built-in compiler (the simplest option), or</li>
<li>The external compiler (the most flexible option, if you are already compiling Tailwind CSS outside Pinegrow).</li>
</ul>
`, 'Would you like to use the external build process?', 'Cancel', 'Use the built-in compiler', null, function() {
                //built in
                done && done();
            }, 'Use the external compiler', function() {
                //external
                _this.default_compiler = 'external';
                _this.setDefaultSettings();
                _this.setSettings(_this.settings, true); //save settings

                _this.notifyAboutMissingSource();
                done && done();
            })
        } else {
            done && done();
        }
    }

    notifyAboutMissingSource() {
        pinegrow.showAlert(`<p>Source CSS file for the external Tailwind CSS build process needs to be selected. Please open Tailwind CSS Compiler options and update the information.</p>`, 'Tailwind CSS Source file is not set', 'Cancel', 'Open Tailwind CSS options...', null, function() {
            var skill = pgDmStore.getDesignSkill('tailwind');
            skill && skill.openDialog();
        });
    }

    onProjectLoaded(project) {
        super.onProjectLoaded(project);
        this.default_compiler = 'jit';
    }

    setDefaultSettings() {
        /*
        var p = pinegrow.getCurrentProject();
        var mode = 'jit';
        if(p) {
            if(crsaIsFileExist())
        }*/
        this.settings = {
            mode: this.default_compiler || 'jit',
            external_source_url: null
        }
    }

    activatedOnPage(page) {
        this.class_tracker.enable();
        this.class_tracker.inspectPage(page);
    }

    addSkills() {
        var _this = this;

        var colors_skill = new PgDmDesignSkillColorsForTailwind(this.colors);
        this.addSkill( colors_skill );

        this.fonts_skill = new PgDmTailwindFontsSkill();
        this.addSkill( this.fonts_skill );

        this.bck_skill = new PgDmDesignSkillBackground(colors_skill, 'TW');
        this.bck_skill.onGetBackgroundColorCSS = function(val) {
            if(val) {
                return `body { @apply bg-${val}; }
`;
            } else {
                return '';
            }
        }

        this.addSkill( this.bck_skill );
        this.options_skill = new PgDmDesignSkillTailwindOptions(2, this);

       // if(isApp()) {
        this.addSkill(this.options_skill);
       // }

        //this.addSkill( new PgDmDesignSkillFonts(false) );
        //this.addSkill( new PgDmDesignSkillFonts(true) );
        //this.addSkill( new PgDmDesignSkillBackground() );
    }

    getActiveFonts() {
        return this.fonts_skill.getActiveFonts();
    }

    canEditColors() {
        return true;
    }

    supportsColorShades() {
        return true;
    }

    getEngine(done) {
        var _this = this;
        if(!this.engine) {
            this.engine = new PgDmTailwindEngine(function(engine) {
                _this.on_engine_ready_funcs.forEach(function(func) {
                    func(engine);
                })
                _this.on_engine_ready_funcs = [];
                done && done(engine)
            }, this);
        } else if(this.engine.isReady()) {
            done(this.engine);
        } else {
            _this.on_engine_ready_funcs.push(done);
        }
    }

    onClassesChanged() {

        if(this.framework && this.framework.twIsUsingExternalBuild()) {
            return;
        }

        var _this = this;
        this.getEngine(function (engine) {
            engine.jit && engine.update();
        })
    }

    hasOnlyDefaultColors() {
        var defs = ['gray', 'red', 'yellow', 'green', 'blue', 'indigo', 'purple', 'pink'];
        if(defs.length !== this.colors.getNumberOfColors()) return false;
        var defs_c = 0;
        var _this = this;
        defs.forEach(function(cn) {
            if(_this.colors.getColorObject(cn)) {
                defs_c++;
            }
        })
        if(defs_c === defs.length) return true;
        return false;
    }

    onBeforeSurpriseMe(done) {
        var _this = this;

        var project = pinegrow.getCurrentProject();

        if(this.colors.getNumberOfUnlockedColors() === 0 && this.hasOnlyDefaultColors() && !project.twAskedAboutColors) {
            project.twAskedAboutColors = true;
            pinegrow.showAlert(`<p>Default Tailwind CSS colors are locked by default because it makes no sense to change Red into blue, for example.</p><p>You can either unlock some of the default colors, or you can add additional colors and use them with Surprise Me. Such colors should have names like <code>primary</code>, <code>secondary</code>, <code>color3</code> and <code>color4</code> (you can rename the colors at any time).</p><p>Would you like to add these additional colors now and use them with Surprise Me?</p>`,
                'All colors are locked',
                `No, don't change colors`,
                `Yes, add colors`,
                done, function() {
                    _this.addCustomizableColors();
                    done();
                }
                )
        } else {
            done();
        }
    }

    onGetDesignThumbHTML(node, htmls) {
        htmls.fonts = this.fonts_skill.getDesignThumbHTML(node);
    }

    addCustomizableColors() {
        this.colors.add('primary', 'rgba(107,114,128,1)', false, {
            '50': 'rgba(249,250,251,1)',
            '100': 'rgba(243,244,246,1)',
            '200': 'rgba(229,231,235,1)',
            '300': 'rgba(209,213,219,1)',
            '400': 'rgba(156,163,175,1)',
            '500': 'rgba(107,114,128,1)',
            '600': 'rgba(75,85,99,1)',
            '700': 'rgba(55,65,81,1)',
            '800': 'rgba(31,41,55,1)',
            '900': 'rgba(17,24,39,1)'
        }, true);

        this.colors.add('secondary', 'rgba(107,114,128,1)', false, {
            '50': 'rgba(249,250,251,1)',
            '100': 'rgba(243,244,246,1)',
            '200': 'rgba(229,231,235,1)',
            '300': 'rgba(209,213,219,1)',
            '400': 'rgba(156,163,175,1)',
            '500': 'rgba(107,114,128,1)',
            '600': 'rgba(75,85,99,1)',
            '700': 'rgba(55,65,81,1)',
            '800': 'rgba(31,41,55,1)',
            '900': 'rgba(17,24,39,1)'
        }, true);

        this.colors.add('color3', 'rgba(107,114,128,1)', false, {
            '50': 'rgba(249,250,251,1)',
            '100': 'rgba(243,244,246,1)',
            '200': 'rgba(229,231,235,1)',
            '300': 'rgba(209,213,219,1)',
            '400': 'rgba(156,163,175,1)',
            '500': 'rgba(107,114,128,1)',
            '600': 'rgba(75,85,99,1)',
            '700': 'rgba(55,65,81,1)',
            '800': 'rgba(31,41,55,1)',
            '900': 'rgba(17,24,39,1)'
        }, true);

        this.colors.add('color4', 'rgba(107,114,128,1)', false, {
            '50': 'rgba(249,250,251,1)',
            '100': 'rgba(243,244,246,1)',
            '200': 'rgba(229,231,235,1)',
            '300': 'rgba(209,213,219,1)',
            '400': 'rgba(156,163,175,1)',
            '500': 'rgba(107,114,128,1)',
            '600': 'rgba(75,85,99,1)',
            '700': 'rgba(55,65,81,1)',
            '800': 'rgba(31,41,55,1)',
            '900': 'rgba(17,24,39,1)'
        }, true);
    }

    getFontsOptions(base) {
        return this.fonts_skill.getFontsOptions(base);
    }

    getColorOptions(prefix) {
        var rows = this.getColorsForPicker(prefix);
        var r = [];
        for(var i = 0; i < rows.length; i++) {
            let row = rows[i];
            for(var n = 0; n < row.length; n++) {
                r.push(row[n]);
            }
        }
        return r;
    }

    getColorsForPicker(prefix) {
        var rows = [
            [
                {
                    key: prefix + 'current', 'name': 'current',
                    color: 'white',
                    text: 'c'
                },
                {
                    'key': prefix + 'transparent', 'name': 'transparent',
                    'color': 'rgba(0,0,0,0)'
                },
                {
                    'key': prefix + 'black', 'name': 'black',
                    'color': 'black'
                },
                {
                    'key': prefix + 'white', 'name': 'white',
                    'color': 'white'
                }
            ]
        ]
        return this.colors.getColorsForPicker(prefix, rows);
    }

    destroyEngine() {
        this.engine && this.engine.destroy();
        this.engine = null;
    }

    recreateEngine(done) {
        this.destroyEngine();
        this.getEngine(done, this);
    }
}

class PgDmTailwindFontsSkill extends PgDmDesignSkill {
    constructor() {
        super('fonts', 'Fonts');

        this.fixed = ['sans', 'serif'];

        this.max_num = 10;

        var _this = this;

        var bm = new PgToggleButtonMaker();

        for(var i = 1; i <= this.max_num; i++) {
            var fixed_name = i <= this.fixed.length ? this.fixed[i-1] : null;

            this.addControl(`name_${i}`, {
                name: 'Name',
                type: fixed_name ? 'hidden' : 'text',
                default_value: fixed_name,
                start_repeat_item: fixed_name || `Custom font ${i - this.fixed.length}`,
                tw_repeat_item_index: i,
                repeat_item_open: fixed_name !== null,
                add_repeat_item_label: 'Add font family',
                on_create_repeat_item_button: function($c) {
                    var $b = $(`<button class="pg-button" style="margin-left:4px;height: 27px;vertical-align: top;">Manage fonts...</button>`).appendTo($c)
                        .on('click', function(e) {
                            e.preventDefault();
                            pinegrow.fontLibrary.showLibrary();
                        })
                },
                dm_no_lock: 'placeholder',
                on_repeat_item_removed: function(prop_view, $item, fdef) {
                    $item.removeClass('has-value');

                    pinegrow.requireProjectDB(function(db) {
                        db.startBatchUpdates();

                        var i = fdef.tw_repeat_item_index;

                        _this.setParam(`family_${i}`, '');
                        _this.setParam(`name_${i}`, '');

                        for(var n = i + 1; n <= _this.max_num; n++) {
                            const name = _this.getParam(`name_${n}`);
                            const family = _this.getParam(`family_${n}`);

                            if(name || family) {
                                _this.setParam(`family_${n-1}`, family);
                                _this.setParam(`name_${n-1}`, name);
                                _this.setParam(`family_${n}`, '');
                                _this.setParam(`name_${n}`, '');
                            }
                        }

                        db.endBatchUpdates('Remove font family', 'design_panel');

                        pgDmStore.update();
                    })
                }
            })

            this.addControl(`family_${i}`, {
                name: 'Family',
                type: 'select',
                show_empty: true,
                fill_list_on_open: true,
                can_add_items: true,
                options: function() {
                    return pinegrow.fontLibrary.getSelectOptions();
                }
            })

            _this.addControl(`${i}_options_container`,
                {
                    name: 'Use for',
                    type: 'displayhtmlwithlabel',
                    html: '',
                    class: `crsa-field-text-font-options-container crsa-field-text-font-options-container-${i}`,
                    dm_no_lock: 'placeholder'
                }
            )

            for(var h = 1; h <= 6; h++) {
                _this.addControl(`h${h}_${i}`,
                    {
                        name: 'H' + h,
                        type: 'select',
                        toggle_buttons: true,
                        append_to: `.crsa-field-text-font-options-container-${i}`,
                        without_text: true,
                        show_empty: true,
                        dm_no_lock: true,
                        options : [
                            {
                                key : 'h' + h,
                                name : 'H' + h,
                                html : bm.makeText('H' + h, {
                                    attributes: { 'title': 'Use for H' + h } ,
                                    styles: {
                                        padding: '4px 0'
                                    }
                                })
                            }
                        ]
                    }
                )
            }
        }


    }

    getFontsOptions(base) {
        var r = [];
        var mono = false;
        for(let i = 1; i <= this.max_num; i++) {
            const name = i <= this.fixed.length ? this.fixed[i - 1] : this.getParam(`name_${i}`);
            const family = this.getParam(`family_${i}`);
            if(name && (i <= this.fixed.length || family)) {
                r.push({ 'key' : base + name,  'name' : name })
                if(name === 'mono') mono = true;
            }
        }
        if(!mono) {
            r.push({ 'key' : base + 'mono',  'name' : 'mono' })
        }
        return r;
    }

    getActiveFonts() {
        var r = []
        for(let i = 1; i <= this.max_num; i++) {
            const name = i <= this.fixed.length ? this.fixed[i - 1] : this.getParam(`name_${i}`);
            const family = this.getParam(`family_${i}`);
            if(name && family && r.indexOf(family) < 0) {
                r.push(family);
            }
        }
        return r;
    }

    getDesignThumbHTML(designnode) {
        var node = designnode.findOne('dmdesignskill[skill="fonts"]');
        var html = '';
        var c = 0;
        if(node) {
            for (let i = 1; i <= this.max_num; i++) {
                const name = i <= this.fixed.length ? this.fixed[i - 1] : node.getAttribute(`name_${i}`);
                const family = node.getAttribute(`family_${i}`);
                if (name && family && c < 2) {
                    html += `<p style="font-family: ${family.replace(/"/g, "'")};margin: 0px 0px;white-space: nowrap;">${family.replace(/["']/g, "")}</p>`;
                    c++;
                }
            }
        }
        return html;
    }

    onSurpriseMe(done) {
        for(var i = 1; i <= this.max_num; i++) {
            const name = i <= this.fixed.length ? this.fixed[i - 1] : this.getParam(`name_${i}`);
            var key = `family_${i}`;
            if(!this.isLocked(key) && name) {
                var type = 'all';
                if(this.getParam(`h1_${i}`) || this.getParam(`h2_${i}`)) {
                    type = 'h1';
                }
                var f = pinegrow.fontLibrary.getRandomFont(type);
                this.setParam(key, f.family);
            }
        }
        done();
    }

    onGetCSS(part, cs_url) {
        if(part === 'import') {
            var fonts_helper = pgDmStore.design.fonts;
            fonts_helper.clear();

            for(let i = 1; i <= this.max_num; i++) {
                const name = i <= this.fixed.length ? this.fixed[i-1] : this.getParam(`name_${i}`);
                const family = this.getParam(`family_${i}`);

                if(name && family) {
                    var info = {
                        key: name,
                        family: family
                    }
                    fonts_helper.setFont(info);
                }
            }
            const imports = fonts_helper.getImportStatements(cs_url);
            var css = imports.length ? (imports.join("\n") + "\n\n") : '';

            var faces = fonts_helper.getFontFontFaceStatements();
            css += faces.length ? (faces.join("\n\n") + "\n\n") : '';
            return css;

        } else if(part === 'rules') {
            var code = '';

            for(let i = 1; i <= this.max_num; i++) {
                const name = i <= this.fixed.length ? this.fixed[i-1] : this.getParam(`name_${i}`);
                const family = this.getParam(`family_${i}`);

                if(name && family) {
                    var sels = [];
                    for (var h = 1; h <= 6; h++) {
                        if (this.getParam(`h${h}_${i}`)) {
                            sels.push('h' + h);
                        }
                    }
                    if (sels.length) {
                        code += `
${sels.join(',')} {
    @apply font-${name};
}                    
`
                    }
                }
            }
            return code;
        }
        return '';
    }

    onGetConfigObject(part, config) {
        if(part === 'init') {
            config.theme = config.theme || {};
            config.theme.extend = config.theme.extend || {}
            config.theme.extend.fontFamily = config.theme.extend.fontFamily || {};

            for(let i = 1; i <= this.max_num; i++) {
                const name = i <= this.fixed.length ? this.fixed[i-1] : this.getParam(`name_${i}`);
                let family = this.getParam(`family_${i}`);
                if(name && family) {
                    var f = pinegrow.fontLibrary.getFontByFamily(family);
                    if(f) {
                        family = f.getFamilyWithFallback();
                    }
                    config.theme.extend.fontFamily[name] = [family];
                }
            }
        }
    }
}

class PgDmTailwindEngine extends PgDmEngine {

    constructor(on_engine_ready, provider) {
        super(on_engine_ready);

        var _this = this;

        this.design_provider = provider;

        this.external_build = provider.useExternalBuild();
        this.jit = true;
        this.is_ready = false;

        this.done_list = [];

        //used to see if we have to reregister classes in TW
        this.class_cache = {
            colors: '',
            fonts: ''
        }

        this.delayedTasks = new PgDelayedTasks();
        this.twf = pinegrow.findFrameworkByKey('tw');

        this.busy = false;
        this.update_on_done = false;
        var p = pinegrow.getCurrentProject();
        this.custom_cs_url = null;

        if(isApp()) {
            this.fs = require('fs');
        }

        if(this.external_build) {
            if(!provider.settings.external_source_url) {
                provider.notifyAboutMissingSource();
            }
            this.is_ready = true;
            //call async just in case
            setTimeout(function() {
                on_engine_ready && on_engine_ready(_this);
            },1)
            return;
        }

        this.custom_cs_url = p.getAbsoluteUrl('tailwind_theme/tailwind.css');

        var outcs = this.outcs = pinegrow.getStylesheetByUrl(this.custom_cs_url);

        if(!outcs) {
            outcs = this.outcs = PgCSSStylesheet();
            //outcs.fast = true;
            outcs.setUrl(this.custom_cs_url);

            pinegrow.stylesheets.addCrsaStylesheet(outcs);

            outcs.load(this.custom_cs_url, function () {
                _this.is_ready = true;
                on_engine_ready && on_engine_ready(_this);
            }, false, true, true);
        } else {
            _this.is_ready = true;
            on_engine_ready && on_engine_ready(_this);
        }
    }

    getWorker() {
        if(this.design_provider.getTailwindVersion() < 3) {
            if(!this.worker2) {
                this.worker2 = new Worker('frameworks/tailwindcss/resources/tailwind.compiler.v2.js');
            }
            return this.worker2;
        } else {
            if(!this.worker) {
                this.worker = new Worker('frameworks/tailwindcss/resources/tailwind.compiler.js');
            }
            return this.worker;
        }
    }

    activateOnPage(page, done) {
        var _this = this;

        if(!isApp() && page.name === '_info.html') {
            done && done();
            return;
        }
        this.design_provider.activatedOnPage(page);

        if(this.external_build) {
            done && done();
            return;
        }

        if(page.sourceNode) {
            var replacePgel = null;
            if (!page.stylesheets.has(this.outcs)) {
                page.sourceNode.withoutEvents(function () {
                    page.sourceNode.find('link').forEach(function (pgel) {
                        if (pgel.getAttribute('rel') === 'stylesheet') {
                            var href = pgel.getAttribute('href');
                            if (href && href.match(/tailwind(\.min|)/i)) {
                                replacePgel = pgel;
                            }
                        }
                    })

                    page.sourceNode.find('script').forEach(function(pgel) {
                        if((pgel.getAttribute('src') || '').indexOf('cdn.tailwindcss.com') >= 0) {
                            pgel.remove();
                        }
                    })

                    if(!_this.outcs) {
                        pinegrow.showQuickError('Please reopen the project.')
                        done && done();
                    } else {
                        page.stylesheets.attachTo(_this.outcs, null, function () {
                            pinegrow.changeManager.didChange(null, 'Add generated stylesheet', null, page);
                            $('body').trigger('crsa-stylesheets-changed');
                            page.refresh();
                            done && done();
                        }, replacePgel);
                    }
                })
                page.setPageChanged(true);
            } else {
                done && done();
            }
        } else {
            done && done();
        }

        //new page was saved, class tracker detected new classes, need to run update now
        if(page.tw_update_jit_on_activate && this.design_provider.useTailwindJIT()) {
            page.tw_update_jit_on_activate = false;
            this.update();
        }
    }

    getTailwindCS(page) {
        return this.outcs;
        //return page.stylesheets.getByRegExp(/tailwind(\.min)?\.css/i);
    }

    update(done) {
        done && this.done_list.push(done);
        this.delayedTasks.run('update', this.updateTask.bind(this), this.external_build ? 1 : 100);
    }

    updateTask() {
        var _this = this;
        var project = pinegrow.getCurrentProject();

        var page = pinegrow.getSelectedPage();
        if(!page) {
            _this.done_list.forEach(function(f) {
                f();
            })
            _this.done_list = [];
            return;
        }

        if(this.busy && Date.now() - this.busy_time > 4000) {
            this.busy = false;
        }

        if(this.busy) {
            this.update_on_done = true;
            return;
        }

        var generate_wp_css = false;
        var wpf = pinegrow.findFrameworkByKey('wordpress.pinegrow');
        if(wpf) {
            generate_wp_css = wpf.isWPEnabledForProject();
        }

        this.external_build = this.design_provider.useExternalBuild();
        this.jit = this.design_provider.useTailwindJIT();

        var o = {
            css: '',
            config: {},//this.design_provider.getExternalConfig(),
            cs_url: this.custom_cs_url
        }

        o.config.mode = this.jit ? 'jit' : '';

        o.config.customPlugins = {
            typography: true,
                forms: {
                active: true
            },
            aspectRatio: true,
            lineClamp: true
        }

        o.config.corePlugins = {
            preflight: true
        }

        pinegrow.dispatchEvent('on_dm_get_css', null, 'import', o);
        pinegrow.dispatchEvent('on_dm_get_css', null, 'init', o);

        var init_css = o.css;

        var source_css = o.css + `
${this.design_provider.disablePreflight() ? '' : '@tailwind base;'}
@tailwind components;
@tailwind utilities;
`;

        this.wp_css = null;

        if(generate_wp_css) {
            source_css = o.css + `
/* PREFLIGHT START */${this.design_provider.disablePreflight() ? '' : '@tailwind base;'}/* PREFLIGHT END */
@tailwind components;
@tailwind utilities;
`;
        }

        o.css = '';
        pinegrow.dispatchEvent('on_dm_get_css', null, 'rules', o);
        source_css += "\n" + o.css;

        var reregister = false;
        var color_classes = JSON.stringify(o.config.theme.extend.colors);
        var font_classes =  JSON.stringify(o.config.theme.extend.fontFamily);

        if(this.class_cache.colors !== color_classes || this.class_cache.fonts !== font_classes) {
            reregister = true;
            this.class_cache.colors = color_classes;
            this.class_cache.fonts = font_classes;
        }

        if(this.external_build) {
            try {
                this.external_config_path = project.getAbsolutePath('tailwind.config.js');
                this.external_source_file = crsaMakeFileFromUrl(project.getAbsoluteUrl(this.design_provider.settings.external_source_url));

                var js = this.fs.readFileSync(this.external_config_path, {encoding: 'utf8'});

                var js_parser = new PgSimpleJsParser(js);
                /*
                if(!js_parser.setProp('extend.colors', o.config.theme.extend.colors)) {
                    throw 'Unable to set extend.colors';
                }
                if(!js_parser.setProp('extend.fontFamily', o.config.theme.extend.fontFamily)) {
                    throw 'Unable to set extend.fontFamily';
                }

                 */
                var dp_code = `
const pg_colors = ${js_parser.objToString(o.config.theme.extend.colors)}

const pg_fonts = ${js_parser.objToString(o.config.theme.extend.fontFamily)}                
                `
                js = pgUpdateGeneratedSectionContent(js, 'Design Panel', dp_code, 'prepend');

                this.fs.writeFileSync(this.external_config_path, js, {encoding: 'utf8'});


                var source_css = this.fs.readFileSync(this.external_source_file, {encoding: 'utf8'});
                source_css = pgUpdateGeneratedSectionContent(source_css, 'Design Panel Init', init_css, 'prepend');
                source_css = pgUpdateGeneratedSectionContent(source_css, 'Design Panel Extra Rules', o.css, 'append');
                this.fs.writeFileSync(this.external_source_file, source_css, {encoding: 'utf8'});

                var fonts = pgDmStore.design.fonts.getFontUrls();
                _this.addFontsToPageViews(function() {
                    return pinegrow.getAllPages();
                }, fonts);
                //pinegrow.refreshPageChangedStatusForAllPages(true);
            } catch(err) {
                console.warn(`Unable to update tailwind.config.js and source file.`, err);
            }
            //call async just in case
            setTimeout(function() {
                if(_this.update_on_done) {
                    _this.update();
                } else {
                    _this.done_list.forEach(function (f) {
                        f();
                    })
                    _this.done_list = [];
                }
                _this.busy = false;
                _this.update_on_done = false;
            },1)


            return;
        }

        var p = new CrsaProfile();

        this.busy = true;
        this.busy_time = Date.now();

        this.getWorker().onerror = function(error) {
            console.error(error)
            _this.busy = false;
            _this.update_on_done = false;
        }

        this.getWorker().onmessage = function(e) {
            if(e.data.error) {
                console.error(e.data.error);
                _this.busy = false;
                _this.update_on_done = false;
            } else {
                p.show('Compiled TW JIT');
                var cs = _this.getTailwindCS(page);
                if(cs) {
                    /*
                    e.data.css = `*, ::before,
        ::after {
          box-sizing: border-box;
        }` + e.data.css;

                     */

                    cs.ignore_imports_in_apply = true;

                    if(generate_wp_css) {
                        var sans_font = o.config.theme?.extend?.fontFamily?.sans;
                        var wp_base_css = `
*,::before,::after {
  box-sizing: border-box;
}

a {
          color: inherit;
          text-decoration: inherit;
        }
        ol,
        ul,
        menu {
          list-style: none;
          margin: 0;
          padding: 0;
        }
        
        blockquote,
        dl,
        dd,
        h1,
        h2,
        h3,
        h4,
        h5,
        h6,
        hr,
        figure,
        p,
        pre {
          margin: 0;
        }                        
`;
                        if(sans_font) {
                            wp_base_css += `
html {
  font-family: ${sans_font[0]};
}                      
`;
                        }

                        var wp_css = e.data.css;
                        e.data.css = e.data.css.replace('/* PREFLIGHT START */', '').replace('/* PREFLIGHT END */', '')
                        wp_css = wp_css.replace(/\/\* PREFLIGHT START \*\/([\s\S]*)\/\* PREFLIGHT END \*\//m, wp_base_css);

                        _this.wp_css = wp_css;
                    }

                    cs.setSource(e.data.css, function() {
                        p.show('TW JIT Applied to page');
                        //console.log(e.data.css)

                        _this.busy = false;
                        if(_this.update_on_done) {
                            _this.update();
                        } else {
                            _this.done_list.forEach(function(f) {
                                f();
                            })
                            _this.done_list = [];
                        }
                        _this.update_on_done = false;

                        pinegrow.dispatchEvent('on_css_stylesheet_changed', null, cs);

                        pinegrow.selectedElements.repositionMenusDelayed();
                        pinegrow.refreshPageChangedStatusForAllPages(true);
                    });

                    var fonts = pgDmStore.design.fonts.getFontUrls();
                    _this.addFontsToPageViews(cs, fonts);
                } else {
                    _this.busy = false;
                    _this.update_on_done = false;
                }
            }
        }

        pgIsDev() && console.log('Compile with classes', this.twf.classTracker.getList())
        this.getWorker().postMessage({
            css: source_css,
            html: [`<html><body class="${this.twf.classTracker.getList()}"></body></html>`],
            config: o.config
        })

        if(reregister) {
            this.design_provider.onColorListChanged(); //will trigger reregister
        }
/*
        this.jit.render(source_css, [`<div class="${this.twf.classTracker.list}"></div>`], o.config)
            .then(css => {
                p.show('Compiled TW JIT');
                var cs = _this.getTailwindCS(page);
                if(cs) {
                    cs.setSource(css, function() {
                        p.show('TW JIT Applied to page');
                    });
                }
                //console.log(css)
            })
            .catch(error => {
                console.error(error);
            })

 */
    }

    saveCSS() {
        var page = pinegrow.getSelectedPage();
        var cs = this.getTailwindCS(page);
        if(cs && cs.localFile) {
            cs.save(cs.localFile);
            cs.setChanged(false);
        }
        if(this.wp_css) {
            var p = pinegrow.getCurrentProject();
            var wp_file = p.getAbsolutePath('tailwind_theme/tailwind_for_wp_editor.css');
            var update = !crsaIsFileExist(wp_file);
            this.fs.writeFileSync(wp_file, this.wp_css, {encoding: 'utf8'});
            update && pinegrow.refreshCurrentProjectDelayed(true);
        }
    }

    destroy() {
        this.worker && this.worker.terminate();
        this.worker = null;
        this.worker2 && this.worker2.terminate();
        this.worker2 = null;
        this.design_provider = null;
        this.delayedTasks.destroy();
    }
}

class PgClassTracker {
    constructor(design_provider) {
        this.design_provider = design_provider;

        design_provider.class_tracker = this;

        this.changed = false;
        this.has_unsaved_changes = false;
        this.list = '';
        this.map = {}
        this.profile = new CrsaProfile('Class Tracker');
        this.profile.pause();
        this.twf = null;

        this.is_enabled = false;

        this.delayedTasks = new PgDelayedTasks();

        this.temp_classes = [];

        var _this = this;

        pinegrow.addEventHandler('on_page_changes_done', function(page, info, col) {
            var do_inspect = false;

            if(_this.is_enabled) {
                if ((_this.twf && _this.twf.twIsUsingExternalBuild())) {
                    do_inspect = true;
                } else if (pgDmStore.isDesignEnabled()) {
                    var dp = pgDmStore.getActiveDesignProvider();
                    if (dp && dp.useTailwindJIT && dp.useTailwindJIT()) {
                        do_inspect = true;
                    }
                }
            }

            if(do_inspect) {
                col && _this.inspectCollection(col);
                if (_this.changed) {
                    _this.changed = false;
                    _this.design_provider.onClassesChanged();
                }
            }
        })

        pinegrow.addEventHandler('on_fast_dom_class_preview_start', function(page, pgel) {
            _this.temp_classes = [];
        })

        pinegrow.addEventHandler('on_fast_dom_class_preview', function(page, pgel, cls, values) {
            if(!_this.is_enabled) return;
            if(!pgDmStore.isDesignEnabled()) return;

            var changed = false;

            var check = function(cls) {
                if (_this.temp_classes.indexOf(cls) < 0 && !_this.map[cls]) {
                    _this.temp_classes.push(cls);
                    changed = true;
                }
            }

            check(cls)
            values && values.forEach(check)

            if(changed) {
                _this.design_provider.onClassesChanged();
                _this.saveUsedClassesHtmlFileIfUsingExternalBuild();
            }
        })

        pinegrow.addEventHandler('on_fast_dom_class_preview_end', function(page, pgel) {
            _this.temp_classes = [];
        })


        pinegrow.addEventHandler('on_project_loaded', function(pagenull, project) {
            _this.twf = pinegrow.findFrameworkByKey('tw');
            _this.is_enabled = false;
            _this.reset();
            if(!_this.load()) {
                //_this.inspectProject(null, true)
            } else {
                _this.is_enabled = true;
            }
            if(_this.twf && _this.twf.twIsUsingExternalBuild()) {
                _this.is_enabled = true;
            }
        });

        pinegrow.addEventHandler('on_page_saved', function(page, saved_scss, first_save) {
            if(_this.is_enabled) {
                //a new page is saved, added to project, let's index it
                if(first_save) {
                    _this.inspectPage(page);
                    if(_this.changed) {
                        page.tw_update_jit_on_activate = true;
                    }
                }
                _this.has_unsaved_changes && _this.save();
            }
        });

        pinegrow.addEventHandler('on_project_closed', function(pagenull, project) {
            if(isApp()) {
                _this.is_enabled && _this.save(project);
                _this.is_enabled = false;
            }
        });

        pinegrow.addEventHandler('on_before_save_and_close_backend_project', function(pagenull, project) {
            if(_this.has_unsaved_changes) {
                _this.save(); //trigger save so that the file will be saved in the project
            }
        });

        pinegrow.addEventHandler('on_project_scan_begin', function(pagenull, project) {
            _this.useOnProject(project) &&  _this.reset();
        })

        pinegrow.addEventHandler('on_project_scan_page', function(pagenull, project, page) {
            _this.useOnProject(project) && _this.inspectPage(page)
        })

        pinegrow.addEventHandler('on_project_scan_end', function(pagenull, project) {
            //_this.changed && _this.save();
        })
    }

    getList() {
        //console.warn('getList', this.temp_classes);
        if(this.temp_classes.length) {
            return this.list + ' ' + this.temp_classes.join(' ');
        } else {
            return this.list;
        }
    }

    rescanProject() {
        var project = pinegrow.getCurrentProject();
        if(!this.useOnProject(project)) {
            pinegrow.showQuickError('Class tracking is not enabled on this project.');
        } else {
            project.forEachEditableFile(function (page, pageDone, status) {
                pageDone(page, status);

            }, function () {
                //on done
                pinegrow.showQuickMessage('Done.');
            }, 'Looking for used classes in HTML files...')
        }
    }

    enable() {
        this.is_enabled = true;
    }

    useOnProject(p) {
        return this.is_enabled;
    }

    inspect(pgel) {
        this.profile.resume();

        var save = false;
        var _this = this;
        pgel.walkSelfAndChildren(function(node) {
            if(node.isElement) {
                node.getClasses().forEach(function(cls) {
                    if(!_this.map[cls]) {
                        _this.map[cls] = true;
                        _this.list += ' ' + cls;
                        !_this.changed && (_this.changed = true);
                        _this.has_unsaved_changes = true;
                        save = true;
                        //console.log('classTracker - new class ' + cls)
                    }
                })
            }
            return true;
        })

        //save && this.delayedSave();
        if(this.changed) {
            this.saveUsedClassesHtmlFileIfUsingExternalBuild();
        }
        this.profile.pause();
        this.profile.show('Total time');

        /*
        if(!isApp() && this.has_unsaved_changes) {
            pinegrow.getCurrentProject().settingsFileIsChanged('pinegrow.json');
        } */
    }

    saveUsedClassesHtmlFileIfUsingExternalBuild() {
        var _this = this;
        if(this.twf && this.twf.twIsUsingExternalBuild()) {
            this.delayedTasks.run('saveUsedClassesHtml', function() {
                var project = pinegrow.getCurrentProject();
                var html_file = project.getAbsolutePath('_pginfo/used-classes.html');
                require('fs').writeFile(html_file, `<html><body><div class="${_this.list} ${_this.temp_classes.join(' ')}"></div></body></html>`, function () {

                });
            }, 5);
        }
    }

    inspectCollection(col) {
        var _this = this;
        col.forEachElement(function(pgel) {
            _this.inspect(pgel);
        })
    }

    inspectPage(page) {
        this.inspect(page.sourceNode);
    }

    inspectProject(done, silent) {
        var _this = this;

        this.reset();

        var delay = silent ? 100 : 0;
        var p = pinegrow.getCurrentProject();
        if(p) {
            p.forEachEditableFile( function( page, pageDone, status ) {
                //on page
                _this.inspectPage(page);

                //wait a bit
                setTimeout(function() {
                    pageDone(page, status);
                }, delay)
            }, function() {
                //on done
            }, 'Looking for used classes...', null, null, silent)
        } else {
            pinegrow.showQuickMessage('Looking for classes in all open pages...');
            pinegrow.getAllPages().forEach(function(page) {
                _this.inspectPage(page);
            })
            pinegrow.showQuickMessage('Done!');
            done && done();
        }
    }

    reset() {
        this.changed = false;
        this.has_unsaved_changes = false;
        this.list = '';
        this.map = {}
    }

    delayedSave() {
        this.delayedTasks.run('save', this.save.bind(this), 1000);
    }

    save(project) {
        var pi;
        if(project) {
            pi = project.getProjectInfo();
        } else {
            pi = pinegrow.getCurrentProjectInfo();
        }
        if(pi) {
            pi.saveFile('class.tracker.json', JSON.stringify({
                list: this.list,
                map: this.map
            }))
            this.has_unsaved_changes = false;
        }
    }

    load() {
        var pi = pinegrow.getCurrentProjectInfo();
        if(pi) {
            var c = pi.loadFile('class.tracker.json');
            if(c) {
                try {
                    var d = JSON.parse(c);
                    this.list = d.list;
                    this.map = d.map;
                    return true;
                } catch(err) {}
            }
        }
        return false;
    }
}