class PgSimpleJsParser {
    constructor(code) {
        this.code_aware_pairs = [
            ['/*', '*/'],
            ['//', '\n'],
            ['"', '"'],
            ["'", "'"]
        ]
        this.code_aware_triggers = /[\'\"\/]/;
        this.parse(code);
    }

    parse(code) {
        this.tokens = this.split(code, '{');
    }

    getCode() {
        return this.tokens.join('{');
    }

    findTokenIndexForProp(prop, start_idx, end_idx) {
        if(start_idx === undefined) start_idx = 0;
        if(end_idx === undefined) end_idx = this.tokens.length - 1;

        var re = new RegExp('(^|[\\s\\n,])' + pgParserEscapeRegExp(prop) + '\\:\\s*$');
        for(let i = start_idx; i <= end_idx; i++) {
            if(this.tokens[i].match(re)) {
                return i + 1; //the next token is the prop { }
            }
        }
        return -1;
    }

    findEndTokenForProp(idx) {
        if(idx >= 0) {
            var end_idx = idx;
            //find the end of the object
            var i = 0;
            var level = 0;
            var found = false;
            while(true) {
                if(end_idx === this.tokens.length) break;

                var close_idx = this.indexOf(this.tokens[end_idx], '}', i);
                if (close_idx >= 0) {
                    if (level === 0) {
                        //found it
                        found = true;
                        break;
                    } else {
                        level--;
                        i = close_idx;
                    }
                } else {
                    end_idx++;
                    level++;
                }
            }
        }
        if(found) {
            return {
                token_idx: end_idx,
                close_char_idx: close_idx
            }
        }
        return null;
    }

    quote(str) {
        var q = str.indexOf("'") >= 0 ? '"' : "'";
        return q + str + q;
    }

    objToString(obj) {
        if(Array.isArray(obj)) {
            var r = '[';
            for(var i = 0; i < obj.length; i++) {
                r += `${this.objToString(obj[i])},\n`;
            }
            r += ']';
            return r;
        } else if(typeof obj === 'object') {
            var r = '{';
            for(var key in obj) {
                var lab = key;
                if(key.match(/[0-9\s\-]/)) {
                    lab = `'${key}'`;
                }
                r += `${lab}: ${this.objToString(obj[key])},\n`;
            }
            r += '}';
            return r;
        } else if(typeof obj === 'number' || typeof obj === 'string') {
            return this.quote(obj);
        }
    }

    setProp(prop, obj) {

        var obj_code = this.objToString(obj);

        var props = prop.split('.');
        if(props.length === 2) { //extend.color
            var parent_idx = this.findTokenIndexForProp(props[0]);
            var parent_end = this.findEndTokenForProp(parent_idx);
            if(!parent_end) {
                console.log(`Unable to find parent property ${props[0]}`);
                return false;
            }
            prop = props[1];
            var idx = this.findTokenIndexForProp(prop, parent_idx, parent_end.token_idx);
            if(idx < 0) {
                //not found, add it
                var added_code = this.addProp(prop, obj_code, parent_idx);
                parent_end.token_idx += added_code.tokens.length;
                idx = this.findTokenIndexForProp(prop, parent_idx, parent_end.token_idx);
                if(idx < 0) {
                    console.log(`Unable to add property ${props[0]}.${prop} to JS code`);
                    return false;
                } else {
                    return true;
                }
            }
        } else {
            var idx = this.findTokenIndexForProp(prop);
        }
        var end = this.findEndTokenForProp(idx);

        if(end) {
            var end_idx = end.token_idx;
            var close_idx = end.close_char_idx;

            var code = new PgSimpleJsParser(obj_code);

            this.tokens[end_idx] = this.tokens[end_idx].substr(close_idx + 1);
            (end_idx - idx > 0) && this.tokens.splice(idx, end_idx - idx);

            var last = code.tokens.pop();
            this.tokens.splice(idx, 0, ...code.tokens);
            this.tokens[idx + code.tokens.length] = last + this.tokens[idx + code.tokens.length];

            try {
                //check for syntax errors
                var func = new Function(this.getCode());
            } catch(err) {
                console.log('Unable to update JS code', err);
                return false;
            }
            return true;
        }
        return false;
    }

    addProp(prop, value, parent_idx) {
        var code = new PgSimpleJsParser(`${prop}: ${value},`);
        //this.tokens.splice(parent_idx, 0, ...code.tokens);
        var last = code.tokens.pop();
        this.tokens.splice(parent_idx, 0, ...code.tokens);
        this.tokens[parent_idx + code.tokens.length] = last + this.tokens[parent_idx + code.tokens.length];
        return code;
    }

    split(str, split_str) {
        return splitCodeAware(str, split_str, this.code_aware_pairs, this.code_aware_triggers);
    }

    indexOf(str, search, i) {
        return indexOfCodeAware(str, search, i || 0, this.code_aware_pairs, this.code_aware_triggers);
    }
}

class PgSimpleJsParserTest {
    constructor() {
    }

    test1() {
        var code = `module.exports = {
  content: [],
  theme: {
    extend: {},
  },
  plugins: [
    require('@tailwindcss/forms'),
    require('@tailwindcss/typography'),
  ],
}
`;
        var p = new PgSimpleJsParser(code);
        console.log(p.getCode(), p.getCode() === code);

        p.setProp('extend', {
            red: 'red',
            blue: 'blue'
        })

        console.log(p.getCode(), p.getCode() === code);
    }

    test2() {
        var code = `module.exports = {
  content: [],
  theme: {
    extend: {
        colors: {
            pink: 'pink',
            miskolin: 'miskolin'
        }
    },
  },
  plugins: [
    require('@tailwindcss/forms'),
    require('@tailwindcss/typography'),
  ],
}
`;
        var p = new PgSimpleJsParser(code);
        console.log(p.getCode(), p.getCode() === code);

        p.setProp('colors', {
            red: {
                '100': '100'
            }
        })

        console.log(p.getCode(), p.getCode() === code);
    }

    test3() {
        var code = `module.exports = {
  content: [],
  theme: {
    extend: {
    },
  },
  plugins: [
    require('@tailwindcss/forms'),
    require('@tailwindcss/typography'),
  ],
}
`;
        var p = new PgSimpleJsParser(code);
        console.log(p.getCode(), p.getCode() === code);

        p.setProp('extend.colors', {
            red: {
                '100': '100'
            }
        })

        p.setProp('extend.colors', {
            blue: {
                '100': '100'
            }
        })

        console.log(p.getCode(), p.getCode() === code);
    }
}
